import configparser
from typing import Generator, Any
from heaserver.service.db.mongo import MongoManager, Mongo
from heaserver.service import appproperty
from heaserver.service.client import get_property
from heaserver.service.oidcclaimhdrs import SUB
from heaserver.service.heaobjectsupport import PermissionGroup
from heaobject.person import Person
from heaobject.user import NONE_USER, ALL_USERS
from heaobject.root import ShareImpl, Permission, Share
from aiohttp.web import Request
from yarl import URL
import logging
from pathlib import Path
from datetime import datetime
from enum import Enum


KEYCLOAK_QUERY_ADMIN_SECRET = 'KEYCLOAK_QUERY_ADMIN_SECRET'
DEFAULT_CLIENT_ID = 'client-cli'
DEFAULT_REALM = 'hea'
DEFAULT_HOST = 'https://localhost:8444'
DEFAULT_SECRET_FILE = '.secret'
DEFAULT_VERIFY_SSL = True

CONFIG_SECTION = 'Keycloak'
KEYCLOAK_IMAGE = 'jboss/keycloak:15.0.2'


class KeycloakCompatibility(Enum):
    """Keycloak compatibility levels:
        FIFTEEN: APIs prior to version 19. We have only tested with 15.
        NINETEEN: APIs for version 19 and later.
    """
    FIFTEEN = "15"
    NINETEEN = "19"

DEFAULT_KEYCLOAK_COMPATIBILITY = KeycloakCompatibility.FIFTEEN

class KeycloakMongo(Mongo):
    """
    Database object for accessing a keycloak server. It subclasses Mongo so that some user data that Keycloak does not
    support might be stored in Mongo.
    """
    def __init__(self, config: configparser.ConfigParser | None = None,
                 client_id: str | None = DEFAULT_CLIENT_ID,
                 realm: str | None = DEFAULT_REALM,
                 host: str | None = DEFAULT_HOST,
                 secret: str | None = None,
                 secret_file: str | None = DEFAULT_SECRET_FILE,
                 verify_ssl: bool = DEFAULT_VERIFY_SSL,
                 keycloak_compatibility: KeycloakCompatibility | None = KeycloakCompatibility.FIFTEEN):
        """
        Initializes Keycloak access with a configparser object or manually set configuration parameters. For all
        manually set configuration parameters, the empty string is treated the same as None.

        :param config: a configparser.ConfigParser object, which should have a Keycloak section with the following
        properties:

            Realm = the Keycloak realm.
            VerifySSL = whether to verify SSL certificates (defaults to yes).
            Host = the keycloak hostname.
            Secret = the secret for accessing keycloak.
            SecretFile = alternatively, a file with one line containing the secret.
            Compatibility = 15 or 19 denoting Keycloak 15-18 versus >= 19. The default is 15.

        :param client_id: the Keycloak client id. The default is client-cli.
        :param realm: the realm to use if there is no config file or the config file does not specify one. The default
        is hea.
        :param host: the hostname to use if there is no config file or the config file does not specify one. The
        default is localhost.
        :param secret: the secret to use if there is no config file or the config file does not specify one.
        :param secret_file: the path of a file with one line containing the secret if there is no config file or the
        config file does not specify one. There must be either a secret or a secret file.
        :param verify_ssl: whether to verify Keycloak's SSL certificate. The default value is True.
        :param keycloak_compatibility: the compatibility level if there is no config file or the config file does not
        specify one. Defaults to FIFTEEN.
        """
        super().__init__(config)
        if config and CONFIG_SECTION in config:
            _section = config[CONFIG_SECTION]
            self.__realm = str(_section.get('Realm', realm) or DEFAULT_REALM)
            self.__verify_ssl = _section.getboolean('VerifySSL', verify_ssl if verify_ssl is not None else DEFAULT_VERIFY_SSL)
            self.__host = str(_section.get('Host', host) or DEFAULT_HOST)
            _secret = _section.get('Secret', secret)
            self.__secret = str(_secret) if _secret else None
            _secret_file = _section.get('SecretFile', secret_file)
            self.__secret_file = str(_secret_file) if _secret_file else None
            compat = _section.get('Compatibility', None) or keycloak_compatibility or DEFAULT_KEYCLOAK_COMPATIBILITY.value
            self.__keycloak_compatibility = KeycloakCompatibility(compat)
        else:
            self.__realm = str(realm) if realm else DEFAULT_REALM
            self.__verify_ssl = bool(verify_ssl) if verify_ssl is not None else DEFAULT_VERIFY_SSL
            self.__host = str(host) if host else DEFAULT_HOST
            self.__secret = str(secret) if secret else None
            self.__secret_file = str(secret_file) if secret_file else None
            if keycloak_compatibility is not None and not isinstance(keycloak_compatibility, KeycloakCompatibility):
                raise ValueError(f'Keycloak_compatibility must be a KeycloakCompatibility enum value or None but was {keycloak_compatibility}')
            self.__keycloak_compatibility = keycloak_compatibility or DEFAULT_KEYCLOAK_COMPATIBILITY
        logger = logging.getLogger(__name__)
        logger.info('Using Keycloak %s mode', self.__keycloak_compatibility.value)
        if self.__host is None:
            raise ValueError
        logger.debug('host is %s', self.__host)

        self.__client_id = str(client_id) if client_id else DEFAULT_CLIENT_ID

    @property
    def client_id(self) -> str:
        return self.__client_id

    @property
    def realm(self) -> str:
        return self.__realm

    @property
    def host(self) -> str:
        return self.__host

    @property
    def secret(self) -> str | None:
        return self.__secret

    @property
    def secret_file(self) -> str | None:
        return self.__secret_file

    @property
    def verify_ssl(self) -> bool:
        return self.__verify_ssl

    @property
    def keycloak_compatibility(self) -> KeycloakCompatibility:
        return self.__keycloak_compatibility

    async def get_keycloak_access_token(self, request: Request) -> str:
        """
        Request an access token from Keycloak.

        :param request: the HTTP request (request).
        :return: the access token or None if not found.
        """
        session = request.app[appproperty.HEA_CLIENT_SESSION]
        logger = logging.getLogger(__name__)

        if self.keycloak_compatibility == KeycloakCompatibility.FIFTEEN:
            token_url = URL(self.host) / 'auth' / 'realms' / self.realm / 'protocol' / 'openid-connect' / 'token'
        else:
            token_url = URL(self.host) / 'realms' / self.realm / 'protocol' / 'openid-connect' / 'token'
        logger.debug('Requesting new access token using credentials')
        secret_property = await get_property(request.app, KEYCLOAK_QUERY_ADMIN_SECRET)
        if secret_property is not None:
            secret = secret_property.value
            logger.debug('Read secret from registry service')
        elif self.secret_file and (secret_file_path := Path(self.secret_file)).exists():
            secret = secret_file_path.read_text()
            logger.debug('Read secret from file')
        elif self.secret:
            secret = self.secret
            logger.debug('Read secret from config')
        else:
            raise ValueError('No secret defined')
        token_body = {
            'client_secret': secret,
            'client_id': 'admin-cli',
            'grant_type': 'client_credentials'
        }
        logger.debug('Going to verify ssl? %r', self.verify_ssl)
        async with session.post(token_url, data=token_body, verify_ssl=self.verify_ssl) as response_:
            content = await response_.json()
            logging.getLogger(__name__).debug(f'content {content}')
            access_token = content['access_token']
        return access_token

    async def get_users(self, request: Request, access_token: str, params: dict[str, str] | None = None) -> \
        list[Person]:
        """
        Gets a list of users from Keycloak using the '/auth/admin/realms/{realm}/users' REST API call.

        :param access_token: the access token to use (required).
        :param params: any query parameters to add to the users request.
        :return: a list of Person objects, or the empty list if there are none.
        """
        logger = logging.getLogger(__name__)
        session = request.app[appproperty.HEA_CLIENT_SESSION]
        db = request.app[appproperty.HEA_DB]
        if self.keycloak_compatibility == KeycloakCompatibility.FIFTEEN:
            users_url = URL(db.host) / 'auth' / 'admin' / 'realms' / db.realm / 'users'
        else:
            users_url = URL(db.host) / 'admin' / 'realms' / db.realm / 'users'

        if params:
            params_ = {}
            for k, v in params.items():
                match k:
                    case 'name':
                        params_['username'] = v
                    case 'first_name':
                        params_['firstName'] = v
                    case 'last_name':
                        params_['lastName'] = v
                    case _:
                        params_[k] = v
            users_url_ = users_url.with_query(params_)
        else:
            users_url_ = users_url
        logger.debug('Getting users from URL %s', users_url_)
        async with session.get(users_url_,
                               headers={'Authorization': f'Bearer {access_token}', 'cache-control': 'no-cache'},
                               verify_ssl=db.verify_ssl) as response_:
            response_.raise_for_status()
            user_json = await response_.json()
            logger.debug(f'Response was {user_json}')
            persons = []
            for user in user_json:
                person = self.__keycloak_user_to_person(user)
                if not params or all(p for p in params.keys() if getattr(person, p) == params[p]):
                    persons.append(person)
        return persons

    async def get_user(self, request: Request, access_token: str, id_: str) -> Person | None:
        """
        Gets the user from Keycloak with the given id using the '/auth/admin/realms/{realm}/users/{id}' REST API call.

        :param request: the HTTP request (required).
        :param access_token: the access token to use (required).
        :param id_: the user id (required).
        :return: a Person object.
        :raises ClientResponseError if an error occurred or the person was not found.
        """
        logger = logging.getLogger(__name__)
        session = request.app[appproperty.HEA_CLIENT_SESSION]
        db = request.app[appproperty.HEA_DB]
        if self.keycloak_compatibility == KeycloakCompatibility.FIFTEEN:
            user_url = URL(db.host) / 'auth' / 'admin' / 'realms' / db.realm / 'users' / id_
        else:
            user_url = URL(db.host) / 'admin' / 'realms' / db.realm / 'users' / id_
        async with session.get(user_url,
                               headers={'Authorization': f'Bearer {access_token}', 'cache-control': 'no-cache'},
                               verify_ssl=db.verify_ssl) as response_:
            user_json = await response_.json()
            logger.debug(f'Response was {user_json}')
            if 'error' in user_json:
                if user_json['error'] == 'User not found':
                    return None
                else:
                    raise ValueError(user_json['error'])
            return self.__keycloak_user_to_person(user_json)

    def __keycloak_user_to_person(self, user: dict[str, Any]) -> Person:
        """
        Converts a user JSON object from Keycloak to a HEA Person object.

        :param user: a Keycloak user object as a JSON dict.
        :return: a Person object.
        """
        person = Person()
        person.id = user['id']
        person.name = user['username']
        person.first_name = user.get('firstName')
        person.last_name = user.get('lastName')
        person.email = user.get('email')
        person.created = datetime.fromtimestamp(user['createdTimestamp'] / 1000.0)
        person.owner = NONE_USER
        person.source = 'Keycloak';
        share = ShareImpl()
        share.user = ALL_USERS
        share.permissions = [Permission.VIEWER]
        person.shares = [share]
        return person


class KeycloakMongoManager(MongoManager):
    """
    Keycloak database manager object. It subclasses the Mongo database manager so that user data that Keycloak does not
    support can be stored in Mongo.
    """
    def __init__(self, config: configparser.ConfigParser | None = None,
                 client_id: str | None = None,
                 realm: str | None = None,
                 secret: str | None = None,
                 secret_file: str | None = None,
                 verify_ssl: bool = True):
        super().__init__(config)
        self.__client_id = str(client_id) if client_id is not None else DEFAULT_CLIENT_ID
        self.__realm = str(realm) if realm is not None else DEFAULT_REALM
        self.__secret: str | None = str(secret) if secret is not None else None
        self.__secret_file: str | None = str(secret_file) if secret_file is not None else DEFAULT_SECRET_FILE
        self.__verify_ssl: bool = bool(verify_ssl)
        self.__keycloak_external_url: str | None = None

    @property
    def client_id(self) -> str:
        return self.__client_id

    @client_id.setter
    def client_id(self, client_id: str):
        self.__client_id = str(client_id) if client_id is not None else DEFAULT_CLIENT_ID

    @property
    def realm(self) -> str:
        return self.__realm

    @realm.setter
    def realm(self, realm: str):
        self.__realm = str(realm) if realm is not None else DEFAULT_REALM

    @property
    def secret(self) -> str | None:
        return self.__secret

    @secret.setter
    def secret(self, secret: str | None):
        self.__secret = str(secret) if secret is not None else None

    @property
    def secret_file(self) -> str | None:
        return self.__secret_file

    @secret_file.setter
    def secret_file(self, secret_file: str | None):
        self.__secret_file = str(secret_file) if secret_file is not None else None

    @property
    def verify_ssl(self) -> bool:
        return self.__verify_ssl

    @verify_ssl.setter
    def verify_ssl(self, verify_ssl: bool):
        self.__verify_ssl = bool(verify_ssl)

    @property
    def keycloak_external_url(self) -> str | None:
        return self.__keycloak_external_url

    def get_database(self) -> KeycloakMongo:
        return KeycloakMongo(config=self.config,
                            client_id=self.client_id,
                            realm=self.realm,
                            host=self.keycloak_external_url,
                            secret=self.secret,
                            secret_file=self.secret_file,
                            verify_ssl=self.verify_ssl)



def _has_permission(request: Request, person: Person) -> bool:
    return request.headers[SUB] == person.owner or person.owner == ALL_USERS


def _shares_with_user(request: Request, person: Person) -> list[Share]:
    return [s for s in person.shares if s.user == request.headers[SUB] and len(
        set(PermissionGroup.GETTER_PERMS.perms).intersection(s.permissions)) > 0]

